/*
 * This file contains functions which are used to
 * compute the function f(v) = exp(digamma(v)).
 * This can be used in a Variational Bayes
 * version of the EM algorithm, which helps to
 * control the overfitting that occurs with the
 * normal EM algorithm.  See Johnson (2007) for
 * more information.
 */

/*
 * This code is part of a small modification to
 * the GIZA++-v2 software, which changes the M
 * step of the EM algorithm slightly.  
 */

#include <assert.h>
#include <math.h>
#include "VarBayes.h"
#include <iostream>
using namespace std;

/*
 * Computes the function f(v) = exp(digamma(v)).
 */
double f(double v) {
  return exp(digamma(v));
}

/* 
 * Computes the digamma function.
 * Code by Mark Johnson:
 * http://www.cog.brown.edu/~mj/code/digamma.c
 */
double digamma(double x) {
  double result = 0, xx, xx2, xx4;
  //assert(x > 0);
  for ( ; x < 7; ++x)
    result -= 1/x;
  x -= 1.0/2.0;
  xx = 1.0/x;
  xx2 = xx*xx;
  xx4 = xx2*xx2;
  result += log(x)+(1./24.)*xx2-(7.0/960.0)*xx4+(31.0/8064.0)*xx4*xx2-(127.0/30720.0)*xx4*xx4;
  return result;
}

/*
 * Test the f and digamma functions.
 * Commented out so as not to conflict
 * with GIZA's actual main() function.
 * To actually run this program, uncomment
 * and recompile separately from GIZA.
 */
/*int main() {
  for (double d; d < 10; d += 0.1) {
    cout << "digamma(" << d << ") = " << digamma(d) 
	 << " and f(" << d << ") = " << f(d) << "\n";
  }
}*/
